@@ -15,7 +15,7 @@ def default(o):  # pylint: disable=method-hidden
         return list(o)
     if isinstance(o, BaseModel):
         return o.to_dict()
-    return o
+    raise TypeError(f"{repr(o)} is not JSON serializable")
 
 
 def dumps(o, default=default, **kwargs):
