@@ -1,8 +1,10 @@
-import logging
+from typing import Any, List, Union, Dict, Type
 from time import sleep
 from datetime import datetime
 from bson.objectid import ObjectId, InvalidId
 from pymongo import MongoClient, ReturnDocument
+from pymongo.cursor import Cursor
+from pymongo.database import Database
 from mongomock import MongoClient as MockMongoClient
 from pymongo.errors import ServerSelectionTimeoutError
 from pymongo.uri_parser import parse_uri
@@ -19,19 +21,19 @@ class NoCache:
     NAME = "NoCache"
 
     @staticmethod
-    def has(_key):
+    def has(_key) -> bool:
         return False
 
     @staticmethod
-    def set(*_args, **_kwargs):
+    def set(*_args, **_kwargs) -> bool:
         return False
 
     @staticmethod
-    def get(_key):
+    def get(_key) -> Any:
         return None
 
     @staticmethod
-    def delete(_key):
+    def delete(_key) -> bool:
         return False
 
 
@@ -67,18 +69,18 @@ class ObjectsCursor:
         self.cursor = cursor
         self._shard_id = shard_id
 
-    def all(self):
+    def all(self) -> List['StorableModel']:
         return list(self)
 
-    def limit(self, *args, **kwargs):
+    def limit(self, *args, **kwargs) -> 'ObjectsCursor':
         self.cursor.limit(*args, **kwargs)
         return self
 
-    def skip(self, *args, **kwargs):
+    def skip(self, *args, **kwargs) -> 'ObjectsCursor':
         self.cursor.skip(*args, **kwargs)
         return self
 
-    def sort(self, *args, **kwargs):
+    def sort(self, *args, **kwargs) -> 'ObjectsCursor':
         self.cursor.sort(*args, **kwargs)
         return self
 
@@ -127,10 +129,10 @@ class Shard:
         self.readonly = self._config.get("readonly", False)
 
     @property
-    def _dbname(self):
+    def _dbname(self) -> str:
         return parse_uri(self._config["uri"]).get("database")
 
-    def get_client(self):
+    def get_client(self) -> Union[MongoClient, MockMongoClient]:
         if not self._client:
             client_kwargs = self._config.get("pymongo_extra", {})
             if self._mock:
@@ -139,23 +141,23 @@ class Shard:
                 self._client = MongoClient(self._config["uri"], **client_kwargs)
         return self._client
 
-    def reset_conn(self):
+    def reset_conn(self) -> None:
         self._client = None
         self._conn = None
 
-    def init_conn(self):
+    def init_conn(self) -> None:
         ctx.log.info("creating a new mongo connection")
         client = self.get_client()
         self._conn = client[self._dbname]
 
     @property
-    def conn(self):
+    def conn(self) -> 'Database':
         if self._conn is None:
             self.init_conn()
         return self._conn
 
     @intercept_pymongo_errors
-    def get_obj(self, ctor, collection, query):
+    def get_obj(self, ctor: Type['StorableModel'], collection: str, query: Dict) -> Union['StorableModel', None]:
         if not isinstance(query, dict):
             try:
                 query = {"_id": ObjectId(query)}
@@ -171,41 +173,49 @@ class Shard:
         return None
 
     @intercept_pymongo_errors
-    def get_obj_id(self, collection, query):
-        return self.conn[collection].find_one(
+    def get_obj_id(self, collection: str, query: Dict) -> Union[ObjectId, None]:
+        obj = self.conn[collection].find_one(
             query, projection=(), session=self._session
-        )["_id"]
+        )
+        if obj:
+            return obj["_id"]
+        return None
 
     @intercept_pymongo_errors
-    def get_objs(self, ctor, collection, query, **kwargs):
+    def get_objs(self, ctor: Type['StorableModel'], collection: str, query: Dict, **kwargs) -> ObjectsCursor:
         if self._session:
             kwargs["session"] = self._session
         cursor = self.conn[collection].find(query, **kwargs)
         return ObjectsCursor(cursor, ctor, shard_id=self._shard_id)
 
     @intercept_pymongo_errors
-    def get_objs_projected(self, collection, query, projection, **kwargs):
+    def get_objs_projected(self, collection: str, query: Dict, projection: Dict, **kwargs) -> Cursor:
         if self._session:
             kwargs["session"] = self._session
         cursor = self.conn[collection].find(query, projection=projection, **kwargs)
         return cursor
 
     @intercept_pymongo_errors
-    def get_aggregated(self, collection, pipeline, **kwargs):
+    def get_aggregated(self, collection: str, pipeline: List[Dict], **kwargs) -> Cursor:
         if self._session:
             kwargs["session"] = self._session
         cursor = self.conn[collection].aggregate(pipeline, **kwargs)
         return cursor
 
     @intercept_pymongo_errors
-    def count_docs(self, collection, query, **kwargs):
+    def count_docs(self, collection: str, query: dict, **kwargs) -> int:
         return self.conn[collection].count_documents(query, **kwargs)
 
-    def get_objs_by_field_in(self, cls, collection, field, values, **kwargs):
-        return self.get_objs(cls, collection, {field: {"$in": values,},}, **kwargs)
+    def get_objs_by_field_in(self,
+                             ctor: Type['StorableModel'],
+                             collection: str,
+                             field: str,
+                             values: List[Any],
+                             **kwargs) -> ObjectsCursor:
+        return self.get_objs(ctor, collection, {field: {"$in": values}}, **kwargs)
 
     @intercept_pymongo_errors
-    def save_obj(self, obj):
+    def save_obj(self, obj: 'StorableModel') -> None:
         if obj.is_new:
             # object to_dict() method should always return all fields
             data = obj.to_dict(include_restricted=True)
@@ -226,13 +236,13 @@ class Shard:
             )
 
     @intercept_pymongo_errors
-    def delete_obj(self, obj):
+    def delete_obj(self, obj: 'StorableModel') -> None:
         if obj.is_new:
             return
         self.conn[obj.collection].delete_one({"_id": obj._id}, session=self._session)
 
     @intercept_pymongo_errors
-    def find_and_update_obj(self, obj, update, when=None):
+    def find_and_update_obj(self, obj: 'StorableModel', update: Dict, when: Union[Dict, None] = None) -> None:
         query = {"_id": obj._id}
         if when:
             assert "_id" not in when
@@ -246,20 +256,20 @@ class Shard:
         return new_data
 
     @intercept_pymongo_errors
-    def delete_query(self, collection, query):
+    def delete_query(self, collection: str, query: Dict):
         return self.conn[collection].delete_many(query, session=self._session)
 
     @intercept_pymongo_errors
-    def update_query(self, collection, query, update):
+    def update_query(self, collection: str, query: Dict, update: Dict):
         return self.conn[collection].update_many(query, update, session=self._session)
 
     # SESSIONS
     @intercept_pymongo_errors
-    def get_session(self, sid, collection="sessions"):
+    def get_session(self, sid: str, collection: str = "sessions"):
         return self.conn[collection].find_one({"sid": sid})
 
     @intercept_pymongo_errors
-    def update_session(self, sid, data, expiration, collection="sessions"):
+    def update_session(self, sid: str, data: Dict, expiration, collection="sessions"):
         self.conn[collection].update(
             {"sid": sid}, {"sid": sid, "data": data, "expiration": expiration}, True
         )
@@ -290,13 +300,13 @@ class DB:
     )
 
     @staticmethod
-    def parse_shard_cfg(shard_name, shardcfg, extra):
+    def parse_shard_cfg(shard_name: str, shardcfg: Dict, extra: Dict) -> Dict:
         if "uri" not in shardcfg:
             raise ConfigurationError(f"uri is missing for shard {shard_name}")
         local_extra = shardcfg.get("pymongo_extra", {})
         return {"uri": shardcfg["uri"], "pymongo_extra": {**extra, **local_extra}}
 
-    def parse_cfg(self, dbconf):
+    def parse_cfg(self, dbconf: Dict) -> Dict:
         config = {"shards": {}}
         global_extra = dbconf.get("pymongo_extra", {})
         if "meta" not in dbconf:
@@ -309,7 +319,7 @@ class DB:
                 )
         return config
 
-    def configure(self, dbconf, mock=False):
+    def configure(self, dbconf: Dict, mock=False) -> None:
         config = self.parse_cfg(dbconf)
         self.__meta = Shard(config["meta"], mock=mock)
         self.__shards = {}
@@ -326,24 +336,24 @@ class DB:
         self.l2c = NoCache()
 
     @property
-    def rw_shards(self):
+    def rw_shards(self) -> List[str]:
         return [
             shard_id for shard_id, shard in self.shards.items() if not shard.readonly
         ]
 
     @property
-    def meta(self):
+    def meta(self) -> Shard:
         if not self.initialized:
             raise RuntimeError("attempted to use uninitialized DB")
         return self.__meta
 
     @property
-    def shards(self):
+    def shards(self) -> Dict[str, Shard]:
         if not self.initialized:
             raise RuntimeError("attempted to use uninitialized DB")
         return self.__shards
 
-    def get_shard(self, shard_id):
+    def get_shard(self, shard_id: str) -> Shard:
         if shard_id not in self.shards:
             raise InvalidShardId(shard_id)
         return self.shards[shard_id]
@@ -375,3 +385,5 @@ class DB:
 
 
 db = DB()
+
+from .models.storable_model import StorableModel
