@@ -23,4 +23,12 @@ def save_required(func):
         if this.is_new:
             raise ObjectSaveRequired("This object must be saved first")
         return func(*args, **kwargs)
+
     return wrapper
+
+
+def pick_rw_shard():
+    from .db import db
+    from random import choice
+
+    return choice(db.rw_shards)
